#include <api/client.h>

#include <core/net/error.h>
#include <core/net/http/client.h>
#include <core/net/http/content_type.h>
#include <core/net/http/response.h>

#include <iostream>

namespace http = core::net::http;
namespace net = core::net;

using namespace api;
using namespace std;

Client::Client() :
    cancelled_(false) {
}

namespace {
static QString readText(QXmlStreamReader& xml)
{
    xml.readNext();

    if (xml.tokenType() != QXmlStreamReader::Characters) {
        return QString();
    }

    return xml.text().toString();
}
}

void Client::getSpiegelFeed(const string &path, QXmlStreamReader &reader) {
    // Create a new HTTP client
    auto client = http::make_client();

    // Start building the request configuration
    http::Request::Configuration configuration;

    // Build the URI from its components
    string fullPath = "www.spiegel.de";
    if (!path.empty())
        fullPath += "/" + path;
    net::Uri uri = net::make_uri(fullPath, { "index.rss"} , {});
    configuration.uri = client->uri_to_string(uri);

    cerr << "querying: " << configuration.uri << endl;

    // Give out a user agent string
    configuration.header.add("User-Agent", "spiegel-scope 0.1;");

    // Build a HTTP request object from our configuration
    auto request = client->head(configuration);

    try {
        // Synchronously make the HTTP request
        // We bind the cancellable callback to #progress_report
        auto response = request->execute(
                    bind(&Client::progress_report, this, placeholders::_1));

        // Check that we got a sensible HTTP status code
        if (response.status != http::Status::ok) {
            throw domain_error(response.body);
        }
        // Parse the Xml from the response
        reader.addData(response.body.c_str());
    } catch (net::Error &) {
    }
}

Client::NewsList Client::allNews(const string &prefix, const string &department, const string &) {
    QXmlStreamReader root;

    string path;

    if (prefix.empty()) {
        path = department;
    } else {
        if (department.empty()) {
            path = prefix;
        } else {
            path = prefix + "/" + department;
        }
    }

    getSpiegelFeed(path, root);

    NewsList list;

    while (!root.atEnd() && !root.hasError()) {
        QXmlStreamReader::TokenType token = root.readNext();

        /* If token is just StartDocument, we'll go to next.*/
        if (token == QXmlStreamReader::StartDocument) {
            continue;
        }

        /* If token is StartElement, we'll see if we can read it.*/
        if (token == QXmlStreamReader::StartElement) {
            if (root.name() == "item") {
                root.readNext();
                NewsItem item;
                while (!root.atEnd() && !(root.isEndElement() && root.name() == "item"))
                {
                    if (root.isEndElement()) {
                        root.readNext();
                        continue;
                    }
                    if (root.name() == "title") {
                        item.title = readText(root).toStdString();
                    } else if (root.name() == "link") {
                        item.link = readText(root).toStdString();
                    } else if (root.name() == "description") {
                        item.description = readText(root).toStdString();
                    } else if (root.name() == "category") {
                        item.category = readText(root).toStdString();
                    } else if (root.name() == "enclosure") {
                        QXmlStreamAttributes atts = root.attributes();
                        if (atts.hasAttribute("type") && atts.value("type") == "image/jpeg" && atts.hasAttribute("url")) {
                            item.image = atts.value("url").toString().toStdString();
                        }
                    }
                    root.readNext();
                }
                if ( item.category.length() == 0) {
                    item.category = "Spiegel";
                }
                list.emplace_back(item);
            }
        }
    }
    return list;
}

http::Request::Progress::Next Client::progress_report(
        const http::Request::Progress&) {

    return cancelled_ ?
                http::Request::Progress::Next::abort_operation :
                http::Request::Progress::Next::continue_operation;
}

void Client::cancel() {
    cancelled_ = true;
}
