#include <boost/algorithm/string/trim.hpp>

#include <scope/localization.h>
#include <scope/query.h>

#include <unity/scopes/Annotation.h>
#include <unity/scopes/CategorisedResult.h>
#include <unity/scopes/CategoryRenderer.h>
#include <unity/scopes/QueryBase.h>
#include <unity/scopes/SearchReply.h>

#include <iomanip>
#include <sstream>

namespace sc = unity::scopes;
namespace alg = boost::algorithm;

using namespace std;
using namespace api;
using namespace scope;


const static string NEWS_RENDERER =
        R"(
{
        "schema-version": 1,
        "template": {
        "category-layout": "grid",
        "card-layout": "horizontal",
        "card-size": "small"
        },
        "components": {
        "title": "title",
        "art" : {
        "field": "art"
        },
        "subtitle": "subtitle"
        }
        }
        )";


Query::Query(const sc::CannedQuery &query, const sc::SearchMetadata &metadata) :
    sc::SearchQueryBase(query, metadata) {
}

void Query::cancelled() {
    client_.cancel();
}

void Query::run(sc::SearchReplyProxy const& reply) {
    try {
        const sc::CannedQuery &query(sc::SearchQueryBase::query());
        // Trim the query string of whitespace
        //string query_string = alg::trim_copy(query.query_string());
        sc::VariantMap config = settings();
        cerr << "got config" << endl;
        if (!config.empty() && config["edition"].get_bool()) {
            germanVersion(reply, query);
        } else {
            internationalVersion(reply, query);
        }
    } catch (domain_error &e) {
        // Handle exceptions being thrown by the client API
        cerr << "EXCEPTION" << endl;
        cerr << e.what() << endl;
        reply->error(current_exception());
    }
}

void Query::internationalVersion(const unity::scopes::SearchReplyProxy &reply,
                                 const unity::scopes::CannedQuery &query) {
    sc::Department::SPtr newest = move(sc::Department::create("", query, "All"));
    sc::DepartmentList depList({ move(sc::Department::create("germany", query, "Germany News")),
                                 move(sc::Department::create("europe", query, "European News")),
                                 move(sc::Department::create("world", query, "World News")),
                                 move(sc::Department::create("business", query, "Business News")),
                                 move(sc::Department::create("zeitgeist", query, "Zeitgeist"))
                               });
    newest->set_subdepartments(depList);
    reply->register_departments(newest);

    Client::NewsList list = client_.allNews("international", query.department_id(), "");
    sc::Category::SCPtr cat = reply->register_category("unknown", "", "", sc::CategoryRenderer(NEWS_RENDERER));
    for (const auto &news : list) {
        if (!pushResult(reply, cat, news))
            return;
    }
}

void Query::germanVersion(const unity::scopes::SearchReplyProxy &reply,
                          const unity::scopes::CannedQuery &query) {
    sc::Department::SPtr newest = move(sc::Department::create("", query, "Neueste"));
    sc::DepartmentList depList({ move(sc::Department::create("politik", query, "Politik")),
                                 move(sc::Department::create("wirtschaft", query, "Wirtschaft")),
                                 move(sc::Department::create("panorama", query, "Panorama")),
                                 move(sc::Department::create("sport", query, "Sport")),
                                 move(sc::Department::create("kultur", query, "Kultur")),
                                 move(sc::Department::create("netzwelt", query, "Netzwelt")),
                                 move(sc::Department::create("wissenschaft", query, "Wissenschaft")),
                                 move(sc::Department::create("gesundheit", query, "Gesundheit")),
                                 move(sc::Department::create("karriere", query, "Karriere")),
                                 move(sc::Department::create("reise", query, "Reise")),
                                 move(sc::Department::create("auto", query, "Auto"))
                               });
    newest->set_subdepartments(depList);
    reply->register_departments(newest);

    Client::NewsList list;
    if (query.department_id().empty()) {
        list = client_.allNews("", "schlagzeilen", "");
    } else {
        list = client_.allNews("", query.department_id(), "");
    }

    std::set<std::string> returned_categories;
    for (const auto &r : list) {
        returned_categories.insert(r.category);
    }

    std::map<std::string, CategoryBuffer> display_categories;

    if (query.department_id().empty()) {
        unity::scopes::utility::BufferedResultForwarder::SPtr next_forwarder = 0;
        CategoryBuffer buf;
        for (sc::DepartmentList::reverse_iterator it = depList.rbegin(); it != depList.rend(); ++it) {
            std::shared_ptr<sc::Department const> dep = *it;
            if (returned_categories.find(dep->label()) == returned_categories.end())
                continue;
            buf.category = reply->register_category(dep->label(), dep->label(), "", sc::CategoryRenderer(NEWS_RENDERER));
            next_forwarder = std::make_shared<unity::scopes::utility::BufferedResultForwarder>(reply, next_forwarder);
            buf.buffer = next_forwarder;
            display_categories[dep->label()] = buf;
        }
        
        for (const auto &news : list) {
            unity::scopes::utility::BufferedResultForwarder::SPtr repl = 0;
            sc::Category::SCPtr cat = 0;
            if (display_categories.find(news.category) != display_categories.end()) {
                repl = display_categories[news.category].buffer;
                cat = display_categories[news.category].category;
            } else {
                continue;
            }

            sc::CategorisedResult res(cat);
            res.set_title(news.title);
            res.set_uri(news.link);
            res.set_art(news.image);
            res["subtitle"] = news.description;
            res["description"] = news.description;
            repl->push(res);
        }
    } else {
        sc::Category::SCPtr cat = reply->register_category("unknown", "", "", sc::CategoryRenderer(NEWS_RENDERER));
        for (const auto &news : list) {
            if (!pushResult(reply, cat, news))
                return;
        }
    }
}

bool Query::pushResult(const sc::SearchReplyProxy &reply,
                       sc::Category::SCPtr category,
                       const Client::NewsItem &news) {
    sc::CategorisedResult res(category);
    res.set_title(news.title);
    res.set_uri(news.link);
    res.set_art(news.image);
    res["subtitle"] = news.description;
    res["description"] = news.description;
    return reply->push(res);
}
